module Extract where

{- type-directed extraction of Haskell programs with a lot of unsafeCoerce

Examples:
---------

MiniAgda

  data Vec (A : Set) : Nat -> Set
  { vnil  : Vec A zero
  ; vcons : [n : Nat] -> (head : A) -> (tail : Vec A n) -> Vec A (suc n) 
  }

  fun length : [A : Set] -> [n : Nat] -> Vec A n -> <n : Nat>
  { length .A .zero    (vnil A)         = zero
  ; length .A .(suc n) (vcons A n a as) = suc (length A n as)
  } 

Haskell

  {-# LANGUAGE NoImplicitPrelude #-}
  module Main where
  import qualified Text.Show as Show

  data Vec (a :: *) 
    = Vec_vnil 
    | Vec_vcons { vec_head :: a , vec_tail :: Vec a }
      deriving Show.Show

  length :: forall a. Vec a -> Nat
  length  Vec_vnil        = Nat_zero
  length (Vec_vcons a as) = Nat_suc (length as)

Components:
-----------

Translation from MiniAgda identifiers to Haskell identifiers

-}

import Char

import Control.Applicative
import Control.Monad
import Control.Monad.Error
import Control.Monad.Reader
import Control.Monad.Writer
import Control.Monad.State

import Data.Map (Map)
import qualified Data.Map as Map

import qualified Language.Haskell.Exts.Syntax as H
import Text.PrettyPrint

import Abstract
import qualified HsSyntax as H
import TraceError

-- translation monad

type Translate = StateT TState (ReaderT TContext (ErrorT TraceError IO))

data TState = TState

initSt :: TState
initSt = TState

data TContext = TContext

initCxt :: TContext
initCxt = TContext

runTranslate :: Translate a -> IO (Either TraceError a)
runTranslate t = runErrorT (runReaderT (evalStateT t initSt) initCxt)

-- translation

translateModule :: [EDeclaration] -> Translate (H.Module)
translateModule ds = do
  hs <- translateDecls ds
  return $ H.mkModule hs

translateDecls :: [EDeclaration] -> Translate [H.Decl]
translateDecls ds = concat <$> mapM translateDecl ds

translateDecl :: EDeclaration -> Translate [H.Decl]
translateDecl d = 
  case d of
    MutualDecl _ ds -> translateDecls ds
    OverrideDecl{} -> fail $ "translateDecls internal error: overrides impossible"
    MutualFunDecl _ _ funs -> translateFuns funs
    FunDecl _ fun -> translateFun fun 
    LetDecl _ ts e -> translateFun $ letToFun ts e
    DataDecl n _ _ _ tel ty cs fields -> translateDataDecl n tel ty cs

translateFuns :: [Fun] -> Translate [H.Decl]
translateFuns funs = concat <$> mapM translateFun funs

translateFun :: Fun -> Translate [H.Decl]
translateFun fun = return []

translateDataDecl :: Name -> Telescope -> Type -> [Constructor] -> Translate [H.Decl]
translateDataDecl n tel ty cs = return []


{-
Name translation

  data names        : check capitalization, identity translation
  constructor names : prefix with Dataname_
  destructor names  : ditto
  type-valued lets  : check capitalization, identity
  type-valued funs  : reject!
  lets              : check lowercase
  funs/cofuns       : check lowercase
-}

hsName :: DefId -> Translate H.Name
hsName id = enter ("error translating identifier " ++ show id) $
  case id of
  (DefId Dat n) -> do
    unless (isUpper $ head n) $ 
      fail $ "data names need to be capitalized"
    return $ H.Ident n
  (DefId (Con co) n) -> do
    dataName <- getDataName n
    return $ H.Ident $ dataName ++ "_" ++ n
  -- lets, funs, cofuns. TODO: type-valued funs!
  (DefId _ n) -> do
     unless (isLower $ head n) $
       fail $ "function names need to start with a lowercase letter"
     return $ H.Ident n

-- getDataName constructorName = return dataNamec
getDataName :: Name -> Translate String
getDataName n = return "DATA" 
